from typing import Optional

from plank.app import Application
from plank.decorator.fastapi import routable
from plank.serving.service import Service
from pydantic import BaseModel


class VersionResponse(BaseModel):
    app_version: str = "0.1.0"
    build_version: str = "2022-08-05.00002"


class BuiltinService(Service):

    def __init__(self, name: str, app: Application, serving_path: Optional[str] = None):
        super().__init__(name=name, serving_path=serving_path)
        self.__build_version = app.build_version
        self.__app_version = app.version

    @routable(path="/version", tags=["default"], methods=["GET"])
    async def version(self) -> VersionResponse:
        return VersionResponse(
            build_version=self.__build_version,
            app_version=self.__app_version
        )



from __future__ import annotations

import inspect
from typing import Callable, Any, Type

from fastapi.responses import Response
from plank.descriptor.action import ActionDescriptor
from plank.server.action import Action
from plank.serving.service import Service
from pydantic import BaseModel


def bounding_if_needed(f: Callable, instance: Any, owner: Type[Any]):
    if f is None:
        return None

    if "." in f.__qualname__ and inspect.isfunction(f):
        # need bounding
        return f.__get__(instance, owner)
    else:
        return f


class RouteActionResponser:

    @property
    def descriptor(self) -> RouteActionDescriptor:
        return self.__descriptor

    @property
    def response_model(self) -> BaseModel:
        return self.__response_model

    def get_response_handler(self, instance, owner):
        return bounding_if_needed(self.__unbound_response_handler, instance=instance, owner=owner)

    def get_response_reverser(self, instance, owner):
        return bounding_if_needed(self.__unbound_response_reverser, instance=instance, owner=owner)

    def __init__(self, descriptor: RouteActionDescriptor):
        self.__descriptor = descriptor
        self.__unbound_response_handler = None
        self.__unbound_response_reverser = None
        self.__response_model = None

    def __call__(self, response_model: Type[BaseModel]) -> Callable[[Callable[[Any], Response]], RouteActionDescriptor]:
        self.__response_model = response_model

        def wrapper(unbound_method: Callable[[Any], Response]):
            self.__unbound_response_handler = unbound_method
            return self.descriptor

        return wrapper

    def reverse(self, unbound_method: Callable[[Any, Type[BaseModel]], Any]) -> RouteActionDescriptor:
        self.__unbound_response_reverser = unbound_method
        return self.descriptor


class RouteActionDescriptor(ActionDescriptor):

    @property
    def response(self) -> RouteActionResponser:
        return self.__responser

    def __init__(self,
                 path: str,
                 end_point: Callable,
                 **kwargs):
        super().__init__(path=path, end_point=end_point, **kwargs)
        self.__responser = RouteActionResponser(descriptor=self)
        self.__unbound_exception_catchers = {}

    def make_action(self, instance: Service, owner: Type[Service]) -> Action:
        from plank.server.fastapi.action import RoutableWrapperAction

        end_point = self.end_point(instance=instance, owner=owner)
        path = self.serving_path(instance=instance, owner=owner)

        # prepare args of RoutableWrapperAction.
        extra_args = self.action_extra_args(instance=instance, owner=owner)
        extra_args["response_model"] = self.__responser.response_model or extra_args.get("response_model")
        service_tags = []
        if isinstance(instance, Service) and instance.name() is not None:
            service_tags = [instance.name().title()]
        extra_args["tags"] = (extra_args.get("tags") or []) + service_tags

        action = RoutableWrapperAction(path=path, end_point=end_point, **extra_args)
        response_handler = self.__responser.get_response_handler(instance, owner)
        response_reverser = self.__responser.get_response_reverser(instance, owner)
        if response_handler is not None:
            action.set_response_handler(response_handler)
        if len(self.__unbound_exception_catchers) > 0:
            for exception_type, unbound_exception_catcher in self.__unbound_exception_catchers.items():
                action.set_exception_catcher(unbound_exception_catcher.__get__(instance, owner),
                                             exception_type=exception_type)
        if response_reverser is not None:
            action.set_response_reverser(response_reverser)
        return action

    def catch(self, *exception_types: Type[Exception]) -> Callable[
        [Callable[[Exception], Response]], RouteActionDescriptor]:
        if len(exception_types) == 0:
            exception_types += [Exception]

        def wrapper(unbound_method: Callable[[Exception], Response]):
            for exception_type in exception_types:
                self.__unbound_exception_catchers[exception_type] = unbound_method
            return self

        return wrapper



from __future__ import annotations
from typing import Callable, Optional, List, Type, TYPE_CHECKING
from pydantic import BaseModel
from plank.decorator.action import action

if TYPE_CHECKING:
    from plank.descriptor.fastapi import RouteActionDescriptor

def routable(path: str,
            name: Optional[str] = None,
            methods: Optional[List[str]] = None,
            tags: Optional[List[str]] = None,
            response_model: Optional[Type[BaseModel]] = None,
            description: Optional[str] = None,
            include_in_schema: Optional[bool] = None)->Callable[[Callable], RouteActionDescriptor]:
    from plank.descriptor.fastapi import RouteActionDescriptor
    return action(
        path=path,
        name=name,
        methods=methods,
        tags=tags,
        response_model=response_model,
        description=description,
        include_in_schema=include_in_schema,
        wrapper_descriptor_type=RouteActionDescriptor
    )


from __future__ import annotations

import inspect
from functools import wraps
from typing import List, Optional, Type, Callable, Any, TYPE_CHECKING

from fastapi.responses import Response
from fastapi.routing import APIRoute
from plank.server.action.serving import ServingAction
from plank.server.action.wrapper import WrapperAction
from plank.serving import Serving
from pydantic import BaseModel

from .interface import Routable

if TYPE_CHECKING:
    pass


class FastAPIRouteAction(ServingAction, Routable):
    def __init__(
            self,
            name: str,
            path: str,
            serving: Serving,
            methods: Optional[List[str]] = None,
            tags: Optional[List[str]] = None,
            response_model: Optional[BaseModel] = None,
            description: Optional[str] = None,
            include_in_schema: Optional[bool] = None
    ):
        super().__init__(path=path, serving=serving)
        self.__name = name
        self.__methods = methods
        self.__include_in_schema = include_in_schema if include_in_schema is not None else True
        self.__tags = tags
        self.__response_model = response_model
        self.__description = description

    def routing_path(self) -> str:
        return self.path

    def name(self) -> str:
        return self.__name

    def methods(self) -> List[str]:
        return self.__methods

    def tags(self) -> List[str]:
        return self.__tags

    def description(self) -> Optional[str]:
        return self.__description

    def response_model(self) -> Optional[Type[BaseModel]]:
        return self.__response_model

    def include_in_schema(self) -> bool:
        return self.__include_in_schema

    def route(self, path_prefix: Optional[str] = None) -> APIRoute:
        return self.get_route(end_point=self.serving.perform, path_prefix=path_prefix)


class RoutableWrapperAction(WrapperAction, Routable):

    def __init__(
            self,
            path: str,
            end_point: Callable,
            name: Optional[str] = None,
            methods: Optional[List[str]] = None,
            tags: Optional[List[str]] = None,
            response_model: Optional[Type[BaseModel]] = None,
            description: Optional[str] = None,
            include_in_schema: Optional[bool] = None,
            response_reverser: Optional[Callable] = None
    ):
        super().__init__(path=path, end_point=end_point, response_reverser=response_reverser)
        self.__name = name or (end_point.__name__ if hasattr(end_point, "__name__") else None)
        self.__methods = methods
        self.__include_in_schema = include_in_schema if include_in_schema is not None else True
        self.__tags = tags
        self.__description = description
        self.__response_model = response_model
        self.__response_handler = None
        self.__exception_catchers = {}

        self.__call__ = wraps(end_point)(lambda *args, **kwargs: end_point(*args, **kwargs))

    def name(self) -> str:
        return self.__name

    def methods(self) -> List[str]:
        return self.__methods

    def tags(self) -> List[str]:
        return self.__tags

    def description(self) -> Optional[str]:
        return self.__description

    def set_exception_catcher(self, exception_catcher: Callable[[Exception], Response],
                              exception_type: Type[Exception]):
        self.__exception_catchers[exception_type] = exception_catcher

    def set_response_handler(self, response_handler: Callable[[Any], Response]):
        self.__response_handler = response_handler

    def response_handler(self, *args, **kwargs):
        sig = inspect.signature(self.__response_handler)
        arguments = sig.bind(*args, **kwargs)
        arguments.apply_defaults()
        return self.__response_handler(**arguments.arguments)

    def set_response_model(self, response_model: BaseModel):
        self.__response_model = response_model

    def response_model(self) -> Optional[Type[BaseModel]]:
        end_point = self.end_point()

        if self.__response_model is not None:
            return self.__response_model

        sig = inspect.signature(end_point)
        if sig.return_annotation == inspect.Signature.empty:
            response_model = None
        else:
            # deal with the return annotation become str by __future__.annotations
            if isinstance(sig.return_annotation, str):
                response_model = end_point.__globals__.get(sig.return_annotation)
            else:
                if not isinstance(sig.return_annotation, Response):
                    response_model = sig.return_annotation

        return response_model

    def include_in_schema(self) -> bool:
        return self.__include_in_schema

    def route(self, path_prefix: Optional[str] = None) -> APIRoute:
        end_point = self.end_point()
        end_point_sig = inspect.signature(end_point)

        # resolve the annotation is str and can't catch in endpoint.__global__
        globals().update({
            parameter.annotation: end_point.__globals__[parameter.annotation]
            for parameter in end_point_sig.parameters.values()
            if isinstance(parameter.annotation, str) and parameter.annotation in end_point.__globals__
        })

        @wraps(end_point)
        async def wrapped_end_point(*args, **kwargs):
            try:
                result = end_point(*args, **kwargs)
                if inspect.isawaitable(result):
                    result = await result
                if self.__response_handler is not None:

                    end_point_request_args = end_point_sig.bind(*args, **kwargs)

                    response_handler_sig = inspect.signature(self.__response_handler)
                    if len(response_handler_sig.parameters) > 1:
                        handled_result = self.__response_handler(result, **end_point_request_args.arguments)
                    elif len(response_handler_sig.parameters) > 0:
                        handled_result = self.__response_handler(result)
                    else:
                        handled_result = self.__response_handler()

                    if inspect.isawaitable(handled_result):
                        return await handled_result
                    else:
                        return handled_result
                else:
                    return result

            except Exception as error:
                exception_type = type(error)
                catcher = self.__exception_catchers.get(exception_type) or self.__exception_catchers.get(Exception)
                if catcher is None:
                    raise error
                else:
                    return catcher(error)

        return self.get_route(end_point=wrapped_end_point, path_prefix=path_prefix)



from typing import Optional

from pydantic import BaseSettings, Field


class SwaggerSettings(BaseSettings):
    title: str = Field(default="${app.name} - Documentations", env="SWAGGER_TITLE")
    description: Optional[str] = Field(default=None, env="SWAGGER_DESCRIPTON")
    path: str = Field(default="/docs", env="SWAGGER_PATH")
    openapi_url: str = Field(default="/swagger/openapi.json", env="SWAGGER_OPENAPI_URL")
    secrets_username: Optional[str] = Field(default=None, env="SWAGGER_SECRETS_USERSNAME")
    secrets_password: Optional[str] = Field(default=None, env="SWAGGER_SECRETS_PASSWORD")



from __future__ import annotations

from typing import Optional, TYPE_CHECKING

from fastapi import FastAPI
from plank import logger
from plank.context import Context
from plank.configuration import Configuration
from plank.server import Server
from plank.support.fastapi.builtin import BuiltinService
from plank.support.fastapi.settings import SwaggerSettings
from plank.support.fastapi.swagger import SwaggerAction
from .action import FastAPIRouteAction
from .interface import Routable

if TYPE_CHECKING:
    from plank.app import Application


class FastAPIServer(Server):
    class Delegate(Server.Delegate):
        def server_did_startup(self, server: FastAPIServer): pass

        def server_did_shutdown(self, server: FastAPIServer): pass

    @classmethod
    def build(cls, configuration: Configuration, delegate: FastAPIServer.Delegate, path_prefix: Optional[str] = None,
                     api_version: Optional[str] = None, **fastapi_arguments) -> FastAPIServer:
        from plank.app import Application
        app = Application(delegate=delegate, configuration=configuration)
        api_version = api_version or configuration.app.build_version
        server = FastAPIServer(application=app, version=api_version, delegate=delegate, path_prefix=path_prefix, **fastapi_arguments)
        return server

    @property
    def build_version(self) -> str:
        return self.application.build_version

    @property
    def include_swagger(self) -> bool:
        return self.__include_swagger

    @property
    def fastapi(self) -> FastAPI:
        return self.__fastapi

    @property
    def swagger_settings(self) -> SwaggerSettings:
        return self.__swagger_settings

    @property
    def api_version(self) -> Optional[str]:
        return self.__api_version

    def __init__(self, application: Application,
                 version: Optional[str] = None,
                 path_prefix: Optional[str] = None,
                 delegate: Optional[FastAPIServer.Delegate] = None,
                 include_swagger: Optional[bool] = None,
                 **fastapi_arguments):
        super().__init__(application=application, delegate=delegate, path_prefix=path_prefix)
        self.__api_version = version
        self.__swagger_settings = SwaggerSettings()
        self.__include_swagger = include_swagger if include_swagger is not None else True

        self.__fastapi = FastAPI(
            docs_url=None,
            redoc_url=None,
            version=f"{version}-{application.configuration.name}",
            **fastapi_arguments
        )

        def startup():
            self.did_startup()

            for path, action in self.actions.items():
                if isinstance(action, Routable):
                    routing_action: Routable = action
                    route = routing_action.route(path_prefix=self.path_prefix)
                    logger.debug(f"Added route: {route} at path: {route.path}")
                    self.__fastapi.routes.append(route)

        def shutdown():
            self.did_shutdown()

        self.__fastapi.router.add_event_handler("startup", startup)
        self.__fastapi.router.add_event_handler("shutdown", shutdown)

    def launch(self, **options):
        super(FastAPIServer, self).launch(**options)

        context = Context.standard()
        context.set("swagger_path", self.__swagger_settings.path)

        self.__fastapi.title = context.reword(self.__swagger_settings.title)
        self.__fastapi.openapi_url = context.reword(self.__swagger_settings.openapi_url)
        self.__fastapi.setup()

        builtin_apis_service = BuiltinService(name=None, app=self.application)
        self.add_action(builtin_apis_service.version)
        if self.__include_swagger:
            action = SwaggerAction(settings=self.swagger_settings)
            self.add_action(action)

    async def __call__(self, scope, receive, send):
        await self.__fastapi(scope=scope, receive=receive, send=send)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['plank',
 'plank.decorator',
 'plank.descriptor',
 'plank.server.fastapi',
 'plank.support',
 'plank.support.fastapi']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.78.0,<0.79.0',
 'plank>=0.1.0,<0.2.0',
 'pydantic>=1.9.2,<2.0.0',
 'uvicorn>=0.17.6,<0.18.0']

setup_kwargs = {
    'name': 'plank-extend-fastapi',
    'version': '0.1.0',
    'description': '',
    'long_description': 'None',
    'author': 'Grady Zhuo',
    'author_email': 'grady@ospark.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import List, Optional, Callable

from fastapi.routing import APIRoute
from plank.utils.path import clearify


class Routable:
    def name(self) -> str:
        raise NotImplementedError

    def routing_path(self) -> str:
        raise NotImplementedError

    def methods(self) -> List[str]:
        raise NotImplementedError

    def tags(self) -> List[str]:
        raise NotImplementedError

    def description(self) -> Optional[str]:
        raise NotImplementedError

    def response_model(self):
        raise NotImplementedError

    def include_in_schema(self) -> bool:
        raise NotImplementedError

    def get_route(self, end_point: Callable, path_prefix: Optional[str] = None) -> APIRoute:
        path = clearify(self.routing_path())
        if path_prefix is not None:
            path_prefix = clearify(path_prefix)
            path = f"{path_prefix}/{path}"
        path = f"/{path}"

        name = self.name() or \
               getattr(end_point, "__name__") if hasattr(end_point, "__name__") else path.split("/")[-1]
        methods = self.methods()
        tags = self.tags()
        response_model = self.response_model()
        description = self.description()
        include_in_schema = self.include_in_schema()
        return APIRoute(path=path, name=name, endpoint=end_point, methods=methods, tags=tags,
                        response_model=response_model, description=description, include_in_schema=include_in_schema)

    def route(self, path_prefix: Optional[str] = None) -> APIRoute:
        raise NotImplementedError()



import secrets
from typing import List, Optional, Callable

from fastapi import Depends, HTTPException
from fastapi.openapi.docs import get_swagger_ui_html
from fastapi.routing import APIRoute
from fastapi.security import HTTPBasic, HTTPBasicCredentials
from plank.context import Context
from plank.server.action import Action
from plank.server.fastapi.action import Routable
from plank.support.fastapi.settings import SwaggerSettings


def check_current_username(secrets_username: Optional[str] = None, secrets_password: Optional[str] = None) -> Callable[
    [Optional[str], Optional[str]], bool]:
    def check(credentials: HTTPBasicCredentials = Depends(HTTPBasic())):
        correct_username = secrets.compare_digest(credentials.username, secrets_username)
        correct_password = secrets.compare_digest(credentials.password, secrets_password)
        return (correct_username and correct_password)

    def no_check():
        return True

    if secrets_username == None:
        return no_check
    return check


class SwaggerAction(Action, Routable):
    def __init__(self, settings: SwaggerSettings):
        self.__settings = settings

    def name(self) -> str:
        return "swagger"

    def __routing_path__(self) -> str:
        return self.__settings.path

    def routing_path(self) -> str:
        return self.__settings.path

    def methods(self) -> List[str]:
        return ["GET"]

    def tags(self) -> List[str]:
        return []

    def description(self) -> Optional[str]:
        defaults = Context.standard()
        return defaults.reword(self.__settings.description)

    def response_model(self):
        return None

    def include_in_schema(self) -> bool:
        return False

    def route(self, path_prefix: Optional[str] = None) -> APIRoute:
        check_func = check_current_username(self.__settings.secrets_username, self.__settings.secrets_password)

        async def end_point(user_passed=Depends(check_func)):
            if not user_passed:
                raise HTTPException(
                    status_code=401,
                    detail='Incorrect email or password',
                    headers={'WWW-Authenticate': 'Basic'},
                )

            defaults = Context.standard()
            return get_swagger_ui_html(
                openapi_url=self.__settings.openapi_url,
                title=defaults.reword(self.__settings.title)
            )

        return self.get_route(end_point, path_prefix=path_prefix)
